import os
import json

from qtharmony.config import THEME_RESOURCES
from qtharmony.core import Loader

from .theme_builder import ThemeBuilder

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from PySide6.QtWidgets import QWidget


class ThemeManager:
    __widgets: list["QWidget"] = []
    __current_theme: str = os.path.join(THEME_RESOURCES, "dark-default.json") 

    @classmethod
    def change_theme(cls, __name: str) -> None:
        cls.__current_theme = cls.get_theme_by_name(__name)
        cls.update()

    @staticmethod
    def get_all_themes() -> dict[str, str]:
        res: dict[str, str] = {}

        for theme in os.listdir(THEME_RESOURCES):
            path = os.path.join(THEME_RESOURCES, theme)
            data = Loader.load_json(path)
            res[data["Info"]["name"]] = path
        
        return res
    
    @classmethod
    def get_theme_by_name(cls, __name: str) -> str:
        data = cls.get_all_themes()

        for theme in data.keys():
            try:
                if theme == __name:
                    return data[theme]
            
            except json.decoder.JSONDecodeError:
                ...
        
        raise FileExistsError(f"There is no such theme: {__name}")

    @classmethod
    def load_theme(cls, __path: str) -> None:
        cls.__current_theme = __path
    
    @classmethod
    def update(cls) -> None:
        data = ThemeBuilder.build_theme(cls.__current_theme)

        for widget in cls.__widgets:
            try:
                if hasattr(widget, "stylesheet"):
                    widget.setStyleSheet(data[widget.__class__.__name__] + widget.stylesheet)  # type: ignore
                
                else:
                    widget.setStyleSheet(data[widget.__class__.__name__])
            
            except KeyError:
                if hasattr(widget, "stylesheet"):
                    widget.setStyleSheet(widget.stylesheet)  # type: ignore

    @classmethod
    def add_widgets(cls, *__widgets: "QWidget") -> None:
        for widget in __widgets:
            cls.__widgets.append(widget)
    
    @classmethod
    def remove_widgets(cls, *__widgets: "QWidget") -> None:
        for widget in __widgets:
            cls.__widgets.remove(widget)
    
    @classmethod
    def get_widgets(cls) -> list["QWidget"]:
        return cls.__widgets



from PySide6.QtWidgets import QPlainTextEdit, QTextEdit
from PySide6.QtCore import Qt

from qtharmony.core import StyleSheetLoader
from qtharmony.core.theme import ThemeManager
from qtharmony.core.sizes import AbstractSize

from .scrollbar import ScrollBar

from typing import Optional, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from PySide6.QtWidgets import QWidget
    from PySide6.QtGui import QFont


class TextBox(QPlainTextEdit):
    def __init__(
            self,
            placed: Optional[str] = None,
            placeholder: Optional[str] = None,
            size: Optional[AbstractSize] = None,
            font: Optional["QFont"] = None,
            is_active: bool = True,
            wrap: bool = False,
            *,
            object_name: str = "text-box",
            stylesheet: Optional[str] = None,
            parent: Optional["QWidget"] = None
    ) -> None:
        super().__init__(parent)
        ThemeManager.add_widgets(self)

        if placed is not None: self.setPlainText(placed)
        if placeholder is not None: self.setPlaceholderText(placeholder)
        if size is not None: size.use(self)
        if font is not None: self.setFont(font)
        self.setDisabled(not is_active)
        if not wrap: self.setLineWrapMode(QPlainTextEdit.LineWrapMode.NoWrap)

        self.setVerticalScrollBar(ScrollBar("vertical"))
        self.setHorizontalScrollBar(ScrollBar("horizontal"))

        self.setObjectName(object_name)
        self.stylesheet = StyleSheetLoader.load_stylesheet(
            __file__, "styles/text_box.css", 
            name=self.__class__.__name__, obj_name=f"QPlainTextEdit#{self.objectName()}",
            stylesheet=stylesheet
        )



from typing import TYPE_CHECKING, Optional

from dataclasses import dataclass
from abc import ABC, abstractmethod

if TYPE_CHECKING:
    from PySide6.QtWidgets import QWidget


@dataclass
class Size:
    width: Optional[int]
    height: Optional[int]

    def ratio(self) -> str:
        if len(self.valid_size()) == 2: 
            return "Size"
        else:
            return [i for i in ("width", "height") 
                    if getattr(self, i) is not None][0].capitalize()

    def valid_size(self) -> tuple[int]:
        return tuple(filter(
            lambda x: isinstance(x, int), (self.width, self.height)
        ))  # type: ignore


class AbstractSize(ABC):
    
    @abstractmethod
    def use(self, widget: "QWidget") -> None:
        ...


class SizeBuilder(AbstractSize):
    def __init__(self, width: Optional[int], height: Optional[int], method) -> None:
        if width is None and height is None:
            raise ValueError("Both arguments are None.")

        self.size: Size = Size(width, height)
        self.method = method

    def use(self, widget: "QWidget") -> None:
        method = f"{self.method}{self.size.ratio()}"

        if hasattr(widget, method):
            getattr(widget, method)(*self.size.valid_size())
        
        else:
            raise ValueError("Invalid widget type.")


class SizeGroup(AbstractSize):
    def __init__(self, *sizes) -> None:
        self.sizes: list[AbstractSize] = [*sizes]
    
    def use(self, widget: "QWidget") -> None:
        for size in self.sizes:
            size.use(widget)


class FixedSize(SizeBuilder):
    def __init__(self, width: Optional[int] = None, height: Optional[int] = None) -> None:
        super().__init__(width, height, "setFixed")


class MaximalSize(SizeBuilder):
    def __init__(self, width: Optional[int] = None, height: Optional[int] = None) -> None:
        super().__init__(width, height, "setMaximum")


class MinimalSize(SizeBuilder):
    def __init__(self, width: Optional[int] = None, height: Optional[int] = None) -> None:
        super().__init__(width, height, "setMinimum")



from typing import Optional, TYPE_CHECKING

from PySide6.QtWidgets import (
    QListWidget, QListWidgetItem, QFrame,
    QHBoxLayout
)
from PySide6.QtCore import Qt

from qtharmony.core import StyleSheetLoader
from qtharmony.core.theme import ThemeManager
from qtharmony.core.sizes import AbstractSize

from .label import Label

if TYPE_CHECKING:
    from PySide6.QtWidgets import QWidget


class WidgetsList(QListWidget):
    """
    now is in development.
    should not use.
    """

    def __init__(
            self, 
            size: Optional[AbstractSize] = None,
            title: Optional[str] = None,
            is_active: bool = True,
            *,
            object_name: str = "widgets-list",
            stylesheet: Optional[str] = None,
            parent: Optional["QWidget"] = None,
    ) -> None:
        super().__init__(parent)
        ThemeManager.add_widgets(self)
    
        if size is not None: size.use(self)
        self.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.setDisabled(not is_active)

        self.setObjectName(object_name)
        self.stylesheet = StyleSheetLoader.load_stylesheet(
            __file__, "styles/widgets_list.css",
            name=self.__class__.__name__, obj_name=f"QListWidget#{self.objectName()}",
            stylesheet=stylesheet
        )

    def add_widget(self, widget) -> None:
        item = QListWidgetItem()
        self.addItem(item)
        self.setItemWidget(item, widget)



from PySide6.QtWidgets import QScrollBar
from PySide6.QtCore import Qt

from qtharmony.core import StyleSheetLoader
from qtharmony.core.sizes import AbstractSize
from qtharmony.core.theme import ThemeManager

from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from PySide6.QtWidgets import QWidget


class ScrollBar(QScrollBar):
    def __init__(
            self,
            __orientation: str,
            size: Optional[AbstractSize] = None,
            *,
            object_name: str = "scroll-bar",
            stylesheet: Optional[str] = None,
            parent: Optional["QWidget"] = None
    ) -> None:
        if __orientation == "horizontal": super().__init__(Qt.Orientation.Horizontal, parent)
        elif __orientation == "vertical": super().__init__(Qt.Orientation.Vertical, parent)

        ThemeManager.add_widgets(self)

        if size is not None: size.use(self)

        self.setObjectName(object_name)
        self.stylesheet = StyleSheetLoader.load_stylesheet(
            __file__, "styles/scrollbar.css", 
            name=self.__class__.__name__, obj_name=f"QScrollBar#{self.objectName()}",
            stylesheet=stylesheet
        )



from PySide6.QtWidgets import QPushButton
from PySide6.QtGui import QIcon
from PySide6.QtCore import QSize

from qtharmony.core import StyleSheetLoader
from qtharmony.core.theme import ThemeManager
from qtharmony.core.sizes import AbstractSize

from typing import Optional, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from PySide6.QtWidgets import QWidget
    from PySide6.QtGui import QFont, QPixmap


class PushButton(QPushButton):
    """
    Custom QPushButton widget for push button functionality.

    Methods:
    - __init__(text: Optional[str] = None, size: tuple[int, int] = (100, 25), 
              font: Optional["QFont"] = None, *, stylesheet: Optional[str] = None, 
              parent: Optional["QWidget"] = None): None
              - Initializes the PushButton widget with optional text, size, font, and stylesheet.

    """

    def __init__(
            self, 
            text: Optional[str] = None,
            size: Optional[AbstractSize] = None,
            font: Optional["QFont"] = None, 
            is_active: bool = True,
            icon: Optional[Union[str, "QIcon", "QPixmap"]] = None,
            icon_size: Optional[tuple[int, int]] = None,
            *,
            object_name: str = "button",
            stylesheet: Optional[str] = None,
            parent: Optional["QWidget"] = None, 
    ) -> None:
        """
        Initializes the PushButton widget with optional text, size, font, and stylesheet.

        Args:
            text (Optional[str], optional): The text displayed on the push button. Defaults to None.
            size (tuple[int, int], optional): The size of the push button widget (width, height). Defaults to (100, 25).
            font (Optional["QFont"], optional): The font used for the push button text. Defaults to None.
            stylesheet (Optional[str], optional): Custom stylesheet for the push button widget. Defaults to None.
            parent (Optional["QWidget"], optional): Parent widget of the push button. Defaults to None.
        """ 

        super().__init__(parent)
        ThemeManager.add_widgets(self)

        if icon is not None: 
            if isinstance(icon, str):
                self.setIcon(QIcon(icon))
            else:
                self.setIcon(icon)
        
        if icon_size is not None:
            self.setIconSize(QSize(*icon_size))

        if font is not None: self.setFont(font)
        self.setDisabled(not is_active)

        self.setObjectName(object_name)
        self.stylesheet = StyleSheetLoader.load_stylesheet(
            __file__, "styles/button.css", 
            name=self.__class__.__name__, obj_name=f"QPushButton#{self.objectName()}",
            stylesheet=stylesheet
        )
        
        if size is not None: size.use(self)

        if text is not None:
            self.setText(text)
        
        if stylesheet is not None:
            self.setStyleSheet(self.styleSheet() + stylesheet)



from PySide6.QtWidgets import QFrame

from qtharmony.core import StyleSheetLoader
from qtharmony.core.theme import ThemeManager
from qtharmony.core.sizes import AbstractSize

from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from PySide6.QtWidgets import QWidget


class Frame(QFrame):

    def __init__(
            self,
            size: Optional[AbstractSize] = None,
            is_active: bool = True,
            *,
            object_name: str = "frame",
            stylesheet: Optional[str] = None,
            parent: Optional["QWidget"] = None
    ) -> None:
        super().__init__(parent)
        ThemeManager.add_widgets(self)

        self.setObjectName(object_name)
        self.stylesheet = StyleSheetLoader.load_stylesheet(
            __file__, "styles/frame.css", 
            name=self.__class__.__name__, obj_name=f"QFrame#{self.objectName()}",
            stylesheet=stylesheet
        )

        self.setDisabled(not is_active)

        if size is not None:
            size.use(self)




from PySide6.QtWidgets import QLabel
from PySide6.QtGui import QPixmap

from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from PySide6.QtGui import QPixmap, QImage


class PictureWidget(QLabel):
    def __init__(self, image: Union["QPixmap", str, "QImage"]) -> None:
        super().__init__()

        self.setPixmap(QPixmap(image))
        self.resize(self.pixmap().size())



from typing import Optional, TYPE_CHECKING
from PySide6.QtWidgets import (
    QGroupBox, QHBoxLayout, QVBoxLayout
)

from qtharmony.core import StyleSheetLoader
from qtharmony.core.theme import ThemeManager 
from qtharmony.core.sizes import AbstractSize

if TYPE_CHECKING:
    from PySide6.QtWidgets import QWidget


def copy_method(source_class, method_name):
    
    def decorator(target_class):
    
        def wrapper(self, *args, **kwargs):
            source_method = getattr(source_class, method_name)
            return source_method(self, *args, **kwargs)
    
        setattr(target_class, method_name, wrapper)
        return target_class
    
    return decorator


class GroupBox(QGroupBox):
    """
    Custom QGroupBox widget for grouping widgets.

    Methods:
    - __init__(title: Optional[str] = None, *, stylesheet: Optional[str] = None, parent: Optional["QWidget"] = None): None
              - Initializes the GroupBox widget with optional title and stylesheet.
    """

    def __init__(
            self,
            title: Optional[str] = None,
            size: Optional[AbstractSize] = None,
            orientation: str = "vertical",
            is_active: bool = True,
            *,
            object_name: str = "group-box",
            stylesheet: Optional[str] = None,
            parent: Optional["QWidget"] = None
    ) -> None:
        """
        Initializes the GroupBox widget with optional title and stylesheet.

        Args:
            title (Optional[str], optional): The title of the group box. Defaults to None.
            stylesheet (Optional[str], optional): Custom stylesheet for the group box widget. Defaults to None.
            parent (Optional["QWidget"], optional): Parent widget of the group box. Defaults to None.
        """

        super().__init__(parent)
        ThemeManager.add_widgets(self)

        self.setObjectName(object_name)
        self.stylesheet = StyleSheetLoader.load_stylesheet(
            __file__, "styles/group_box.css", 
            name=self.__class__.__name__, obj_name=f"QGroupBox#{self.objectName()}",
            stylesheet=stylesheet
        )

        if title is not None: self.setTitle(title)
        if size is not None: size.use(self)

        if orientation == "vertical": self.mainLayout = QVBoxLayout()
        elif orientation == "horizontal": self.mainLayout = QHBoxLayout()
        else: raise ValueError("Invalid orientation. should be use 'vertical' or 'horizontal'.")

        self.setDisabled(not is_active)

        self.setLayout(self.mainLayout)
    
    def addWidget(self, *args, **kwargs) -> None:
        self.mainLayout.addWidget(*args, **kwargs)
    
    def addLayout(self, *args, **kwargs) -> None:
        self.mainLayout.addLayout(*args, **kwargs)
    
    def addItem(self, *args, **kwargs) -> None:
        self.mainLayout.addItem(*args, **kwargs)
